/*
 * leddipswdriver.c
 *   2011/9/5   Version 2.1   for linux-3.0 kernel register_chdrv()
 *   2017/6/8   Version 3.0   for linux-3.0.11 ARMv7
 */

#define DEVNAME "leddrv"
#define LEDDRV_MAJOR   241
#define LEDDRV_DIPSW_BASE 128

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/major.h>
#include <linux/device.h>
#include <linux/gpio.h>
#include <asm/uaccess.h>

/* -----------------------------------
 * 低レベル IOポート入出力関数
 * ----------------------------------- */

const int leddrv_ledgpio[]={816, 817};
const int leddrv_dipswgpio[]={818,819,820,821};

#define MAX_LED_NUM (sizeof(leddrv_ledgpio)/sizeof(int))
#define MAX_DIPSW_NUM (sizeof(leddrv_dipswgpio)/sizeof(int))

/* ================ 初期化 ================== */

static void leddrv_hardware_init(void)
{
	int i;
	int gpio;

	for(i=0; i<MAX_LED_NUM; i++){
		gpio = leddrv_ledgpio[i];
		if(gpio_is_valid(gpio)){
			if(gpio_request(gpio,DEVNAME)){
				printk("%s():%d error. gpio_request(%d) \n",__func__,__LINE__,i);
			}
			gpio_direction_output(gpio, 1);
		}else{
			printk("%s():%d error. gpio_is_valid(%d) is not valid\n",__func__,__LINE__,i);
		}
	}

	for(i=0; i<MAX_DIPSW_NUM; i++){
		gpio = leddrv_dipswgpio[i];
		if(gpio_is_valid(gpio)){
			if(gpio_request(gpio,DEVNAME)){
				printk("%s():%d error. gpio_request(%d) \n",__func__,__LINE__,i);
			}
			gpio_direction_input(gpio);
		}else{
			printk("%s():%d error. gpio_is_valid(%d) is not valid\n",__func__,__LINE__,i);
		}
	}

}

/* ================ 終了 ================== */

static void leddrv_hardware_exit(void)
{
	int i;
	int gpio;
	for(i=0; i<MAX_LED_NUM; i++){
		gpio = leddrv_ledgpio[i];
		if(gpio_is_valid(gpio)){
			gpio_free(gpio);
		}else{
			printk("%s():%d error. gpio_is_valid(%d) is not valid\n",__func__,__LINE__,i);
		}
	}

	for(i=0; i<MAX_DIPSW_NUM; i++){
		gpio = leddrv_dipswgpio[i];
		if(gpio_is_valid(gpio)){
			gpio_free(gpio);
		}else{
			printk("%s():%d error. gpio_is_valid(%d) is not valid\n",__func__,__LINE__,i);
		}
	}
}

/* ================ 出力 ================== */

static void leddrv_led_output(int no, int on)
{
	int gpio;

	if(no>=MAX_LED_NUM)
		return;
	if(no<0)
		return;

	gpio = leddrv_ledgpio[no];

	if(on){
		gpio_set_value(gpio, 0);	/* ON */
	}else{
		gpio_set_value(gpio, 1);	/* OFF */
	}
}

/* ================ 入力 ================== */

static int leddrv_dipsw_input(int no)
{
	int gpio;
	int value;

	if(no>=MAX_DIPSW_NUM)
		return 0;
	if(no<0)
		return 0;

	gpio = leddrv_dipswgpio[no];

	if(gpio_get_value(gpio) == 0){
		value=0;
	}else{
		value=1;
	}

	return value;
}

/* -----------------------------------
 * キャラクタ型ドライバ
 * ファイルオペレーション
 * ----------------------------------- */

static struct cdev led_cdev;
static struct class *led_class;

static struct cdev dipsw_cdev;
static struct class *dipsw_class;


static int leddipsw_open(struct inode *inode, struct file *filp){
	// open()されたmajor,minor番号取得
	int major = MAJOR(inode->i_rdev);
	int minor = MINOR(inode->i_rdev);
	dev_t led_devid = led_cdev.dev;
	dev_t dipsw_devid = dipsw_cdev.dev;

	// private_data に デバイス番号(major/minor)保存
	filp->private_data = (void*)(inode->i_rdev);

	if(major == MAJOR(led_devid)){
		if((minor>=0) && (minor<MAX_LED_NUM)){
			/* major, minor番号はLED */
			return 0;	/* ok */
		}
	}

	if(major == MAJOR(dipsw_devid)){
		if((minor>=0) && (minor<MAX_DIPSW_NUM)){
			/* major, minor番号はDIPSW */
			return 0;	/* ok */
		}
	}

	return -ENODEV;	 /* No such device */
}

static int leddipsw_release(struct inode *inode, struct file *filp){
	return 0;
}

static int leddipsw_read(struct file *filp, char *user_buf,size_t count,loff_t *ppos){
	dev_t devid;
	dev_t dipsw_devid;
	int ret;
	int major;
	int minor;

	devid = (dev_t)(filp->private_data);
	dipsw_devid = dipsw_cdev.dev;
	major = MAJOR(devid);
	minor = MINOR(devid);

	if(major == MAJOR(dipsw_devid)){
		if((minor>=0) && (minor<MAX_DIPSW_NUM)){
			unsigned char buffer[16];  // スタックサイズ=ローカル変数のサイズ=約7kbyte程度なので要注意
			int len,data,dipswno;
			dipswno = minor;
			data = leddrv_dipsw_input(dipswno);
			// 文字列に変換する
			len=sprintf(buffer,"%d\n",data);
			if(len>count)
				len=count;
			// copy_to_user(転送先,転送元,バイト長) で
			// ユーザープロセス空間にデータをコピーする
			ret = copy_to_user(user_buf,buffer,len);
			if(ret<0){
				/* エラー発生時はエラー番号を返す */
				return ret;
			}
			return len;  // 転送バイト数を返す
		}
	}
	return 0;	/* 何もしない */
	// 戻り値は
	//   負:エラー
	//   0 :end-of-file に到達
	//   正:転送したバイト数
	// としなければならない
}


static int leddipsw_write(struct file *filp, const char *user_buf,size_t count,loff_t *ppos){
	dev_t devid;
	dev_t led_devid;
	int ret;
	int major;
	int minor;

	devid = (dev_t)(filp->private_data);
	major = MAJOR(devid);
	minor = MINOR(devid);

	led_devid = led_cdev.dev;

	if(major == MAJOR(led_devid)){
		if((minor>=0) && (minor<MAX_LED_NUM)){
			unsigned char c;
			int ledno;
	
			ledno=minor;
			ret = copy_from_user(&c,user_buf,1);
			if(ret<0){
				/* エラー発生時はエラー番号を返す */
				return ret;
			}
			if(c=='0')
				leddrv_led_output(ledno, 0);
			else if(c=='1')
				leddrv_led_output(ledno, 1);
		}
	}
	return count;
	// 戻り値は
	//   負:エラー
	//   0 :end-of-file に到達
	//   正:転送したバイト数
	// としなければならない
	// write()の時は、ユーザープロセスが書き出ししたかったバイト数(count)を
	// そのまま返すと、ユーザープロセスは大満足する(^-^)
}

static const struct file_operations leddrv_fops = {
	.owner          = THIS_MODULE,
	.read           = leddipsw_read,      /* read()  */
	.write          = leddipsw_write,     /* write() */
	.open           = leddipsw_open,      /* open()  */
	.release        = leddipsw_release,   /* close() */
};

/* -----------------------------------
 * ドライバの ロード / アンロード処理
 * ----------------------------------- */

/* キャラクタ型デバイスドライバの登録 LED */
static int led_device_init(void)
{
	int ret;
	dev_t led_devid;
	int i, major;
	/* 空いているキャラクタ型ドライバの major/minor 番号を探して取得する */
	ret = alloc_chrdev_region(&led_devid,   0, MAX_LED_NUM,   DEVNAME);
	if(ret<0)
		goto out1;
	cdev_init(&led_cdev, &leddrv_fops);
	ret = cdev_add(&led_cdev, led_devid, MAX_LED_NUM);
	if(ret<0)
		goto out2;

	/* /sys/class/led を作る */
	led_class = class_create(THIS_MODULE, "led");
	if (IS_ERR(led_class)) {
		printk(KERN_ERR "Error creating led class.\n");
		ret = PTR_ERR(led_class);
		goto out3;
	}

	/* デバイス  /sys/class/led/led0 ～ led1 を作る */
	major=MAJOR(led_devid);
	for(i=0; i<MAX_LED_NUM; i++){
		device_create(led_class, NULL, MKDEV(major,i), NULL, "led%d", i);
	}

	return 0;

out3:
	cdev_del(&led_cdev);
out2:
	unregister_chrdev_region(led_devid, MAX_LED_NUM);
out1:
	return ret;
}

/* キャラクタ型デバイスドライバの登録 dipsw */
static int dipsw_device_init(void)
{
	int ret;
	dev_t dipsw_devid;
	int i, major;

	/* 空いているキャラクタ型ドライバの major/minor 番号を探して取得する */
	ret = alloc_chrdev_region(&dipsw_devid,   0, MAX_DIPSW_NUM,   DEVNAME);
	if(ret<0)
		goto out1;
	cdev_init(&dipsw_cdev, &leddrv_fops);
	ret = cdev_add(&dipsw_cdev, dipsw_devid, MAX_DIPSW_NUM);
	if(ret<0)
		goto out2;

	/* /sys/class/dipsw を作る */
	dipsw_class = class_create(THIS_MODULE, "dipsw");
	if (IS_ERR(dipsw_class)) {
		printk(KERN_ERR "Error creating dipsw class.\n");
		ret = PTR_ERR(dipsw_class);
		goto out3;
	}

	/* デバイス  /sys/class/dipsw/dipsw0 ～ dipsw3 を作る */
	major=MAJOR(dipsw_devid);
	for(i=0; i<MAX_DIPSW_NUM; i++){
		device_create(dipsw_class, NULL, MKDEV(major,i), NULL, "dipsw%d", i);
	}

	return 0;

out3:
	cdev_del(&dipsw_cdev);
out2:
	unregister_chrdev_region(dipsw_devid, MAX_DIPSW_NUM);
out1:
	return ret;
}

static int leddrv_init(void){
	int ret;

	printk(DEVNAME " hello\n");

	/* ハードウェア的な初期化 */
	leddrv_hardware_init();

	ret = led_device_init();
	if(ret<0)
		return ret;

	ret = dipsw_device_init();
	if(ret<0)
		return ret;

	return 0;
}


static void leddrv_exit(void){
	dev_t led_devid = led_cdev.dev;
	dev_t dipsw_devid = dipsw_cdev.dev;
	int i,major;

	printk(DEVNAME " goodby\n");

	/* LEDの登録解除 */
	major=MAJOR(led_devid);
	for(i=0;i<MAX_LED_NUM;i++)
		device_destroy(led_class, MKDEV(major,i));

	class_destroy(led_class);
	cdev_del(&led_cdev);
	unregister_chrdev_region(led_devid, MAX_LED_NUM);

	/* DIPSWの登録解除 */
	major=MAJOR(dipsw_devid);
	for(i=0;i<MAX_DIPSW_NUM;i++)
		device_destroy(dipsw_class, MKDEV(major,i));

	class_destroy(dipsw_class);
	cdev_del(&dipsw_cdev);
	unregister_chrdev_region(dipsw_devid, MAX_DIPSW_NUM);

	leddrv_hardware_exit();

}

module_init(leddrv_init);
module_exit(leddrv_exit);

MODULE_LICENSE("GPL");
